""" Here we update semantic graph with content that classifier produced"""
from cognee.infrastructure.databases.graph.get_graph_client import get_graph_client, GraphDBType


async def add_classification_nodes(document_id, classification_data):
    graph_client = get_graph_client(GraphDBType.NETWORKX)

    await graph_client.load_graph_from_file()

    data_type = classification_data["data_type"]
    layer_name = classification_data["layer_name"]

    # Create the layer classification node ID
    layer_classification_node_id = f"LLM_LAYER_CLASSIFICATION:{data_type}:{document_id}"

    # Add the node to the graph, unpacking the node data from the dictionary
    await graph_client.add_node(layer_classification_node_id, **classification_data)

    # Link this node to the corresponding document node
    await graph_client.add_edge(document_id, layer_classification_node_id, relationship = "classified_as")

    # Create the detailed classification node ID
    detailed_classification_node_id = f"LLM_CLASSIFICATION:LAYER:{layer_name}:{document_id}"

    # Add the detailed classification node, reusing the same node data
    await graph_client.add_node(detailed_classification_node_id, **classification_data)

    # Link the detailed classification node to the layer classification node
    await graph_client.add_edge(layer_classification_node_id, detailed_classification_node_id, relationship = "contains_analysis")

    return True



class Job:
    def __init__(self, job_id, job_status=None, filename=None):
        self.job_id = job_id
        self.job_status = job_status
        self.filename = filename

    def __str__(self):
        attributes = []
        if self.job_id is not None:
            attributes.append(f"job_id: {self.job_id}")
        if self.job_status is not None:
            attributes.append(f"job_status: {self.job_status}")
        if self.filename is not None:
            attributes.append(f"filename: {self.filename}")
        return "Job(" + ", ".join(attributes) + ")"

    def __repr__(self):
        return self.__str__()



""" This module contains the functions that are used to query the language model. """
import os
import instructor
from openai import OpenAI
import logging
from cognee.shared.data_models import  KnowledgeGraph,  MemorySummary
from cognee.config import Config



config = Config()
config.load()

OPENAI_API_KEY = config.openai_key

aclient = instructor.patch(OpenAI())


# Function to read query prompts from files
def read_query_prompt(filename):
    """Read a query prompt from a file."""
    try:
        with open(filename, "r") as file:
            return file.read()
    except FileNotFoundError:
        logging.info(f"Error: File not found. Attempted to read: %s {filename}")
        logging.info(f"Current working directory: %s {os.getcwd()}")
        return None
    except Exception as e:
        logging.info(f"An error occurred: %s {e}")
        return None


def generate_graph(input) -> KnowledgeGraph:
    """Generate a knowledge graph from a user query."""
    model = "gpt-4-1106-preview"
    user_prompt = f"Use the given format to extract information from the following input: {input}."
    system_prompt = read_query_prompt(
        "cognee/llm/prompts/generate_graph_prompt.txt"
    )

    out = aclient.chat.completions.create(
        model=model,
        messages=[
            {
                "role": "user",
                "content": user_prompt,
            },
            {
                "role": "system",
                "content": system_prompt,
            },
        ],
        response_model=KnowledgeGraph,
    )
    return out


async def generate_summary(input) -> MemorySummary:
    """Generate a summary from a user query."""
    out = aclient.chat.completions.create(
        model=config.model,
        messages=[
            {
                "role": "user",
                "content": f"""Use the given format summarize 
                and reduce the following input: {input}. """,
            },
            {
                "role": "system",
                "content": """You are a top-tier algorithm
                designed for summarizing existing knowledge 
                graphs in structured formats based on a knowledge graph.
                ## 1. Strict Compliance
                Adhere to the rules strictly. 
                Non-compliance will result in termination.
                ## 2. Don't forget your main goal 
                is to reduce the number of nodes in the knowledge graph 
                while preserving the information contained in it.""",
            },
        ],
        response_model=MemorySummary,
    )
    return out


def user_query_to_edges_and_nodes(input: str) -> KnowledgeGraph:
    """Generate a knowledge graph from a user query."""
    system_prompt = read_query_prompt(
        "cognee/llm/prompts/generate_graph_prompt.txt"
    )
    return aclient.chat.completions.create(
        model=config.model,
        messages=[
            {
                "role": "user",
                "content": f"""Use the given format to
                 extract information from the following input: {input}. """,
            },
            {"role": "system", "content": system_prompt},
        ],
        response_model=KnowledgeGraph,
    )



""" This module contains the function to classify a summary of a document. """

import json
import logging

from langchain.prompts import ChatPromptTemplate
from langchain.chains import create_extraction_chain
from langchain.chat_models import ChatOpenAI

from ..config import Config

config = Config()
config.load()
OPENAI_API_KEY = config.openai_key

async def classify_summary(query, document_summaries):
    """Classify the documents based on the query and content."""
    llm = ChatOpenAI(temperature=0, model=config.model)
    prompt_classify = ChatPromptTemplate.from_template(
        """You are a  classifier. Determine what document  
        are relevant for the given query: {query}, 
        Document summaries and ids:{document_summaries}"""
    )
    json_structure = [
        {
            "name": "classifier",
            "description": "Classification",
            "parameters": {
                "type": "object",
                "properties": {
                    "DocumentSummary": {
                        "type": "string",
                        "description": "The summary of the document "
                                       "and the topic it deals with.",
                    },
                    "d_id": {"type": "string", "description": "The id of the document"},
                },
                "required": ["DocumentSummary"],
            },
        }
    ]
    chain_filter = prompt_classify | llm.bind(
        function_call={"name": "classifier"}, functions=json_structure
    )
    classifier_output = await chain_filter.ainvoke(
        {"query": query, "document_summaries": document_summaries}
    )
    arguments_str = classifier_output.additional_kwargs["function_call"]["arguments"]
    logging.info("This is the arguments string %s", arguments_str)
    arguments_dict = json.loads(arguments_str)
    logging.info("Relevant summary is %s", arguments_dict.get("DocumentSummary", None))
    classfier_id = arguments_dict.get("d_id", None)

    logging.info("This is the classifier id %s", classfier_id)

    return classfier_id



""" This module is responsible for converting content to cognitive layers. """
import json
from typing import Type
from pydantic import BaseModel
from cognee.infrastructure.llm.get_llm_client import get_llm_client
from cognee.shared.data_models import KnowledgeGraph
from cognee.utils import async_render_template

async def generate_graph(text_input: str, filename: str, context, response_model: Type[BaseModel]):
    llm_client = get_llm_client()

    formatted_text_input = await async_render_template(filename, context)
    output = await llm_client.acreate_structured_output(text_input, formatted_text_input, response_model)

    context_key = json.dumps(context, sort_keys = True)

    # Returning a dictionary with context as the key and the awaited output as its value
    return {
        context_key: output
    }


if __name__ == "__main__":
    import asyncio
    asyncio.run(generate_graph(text_input="bla", filename = "generate_graph_prompt.txt",context= {
        'layer': 'BLA'
    }, response_model=KnowledgeGraph))




""" This module contains the classifiers for the documents. """

import json
import logging

from langchain.prompts import ChatPromptTemplate
from langchain.chains import create_extraction_chain
from langchain.chat_models import ChatOpenAI

from ..config import Config

config = Config()
config.load()
OPENAI_API_KEY = config.openai_key

async def classify_user_input(query, input_type):
    """ Classify the user input based on the query and input type."""
    llm = ChatOpenAI(temperature=0, model=config.model)
    prompt_classify = ChatPromptTemplate.from_template(
        """You are a classifier. 
        Determine with a True or False if the following input: {query}, 
        is relevant for the following memory category: {input_type}"""
    )
    json_structure = [
        {
            "name": "classifier",
            "description": "Classification",
            "parameters": {
                "type": "object",
                "properties": {
                    "InputClassification": {
                        "type": "boolean",
                        "description": "The classification of the input",
                    }
                },
                "required": ["InputClassification"],
            },
        }
    ]
    chain_filter = prompt_classify | llm.bind(
        function_call={"name": "classifier"}, functions=json_structure
    )
    classifier_output = await chain_filter.ainvoke(
        {"query": query, "input_type": input_type}
    )
    arguments_str = classifier_output.additional_kwargs["function_call"]["arguments"]
    logging.info("This is the arguments string %s", arguments_str)
    arguments_dict = json.loads(arguments_str)
    logging.info("Relevant summary is %s", arguments_dict.get("DocumentSummary", None))
    InputClassification = arguments_dict.get("InputClassification", None)
    logging.info("This is the classification %s", InputClassification)
    return InputClassification



"""This module is used to create the database and tables for the cognitive architecture."""
import logging

logger = logging.getLogger(__name__)

async def main():
    """Runs as a part of startup docker scripts to create the database and tables."""
    from cognee.config import Config
    config = Config()
    config.load()

    from cognee.database.database_manager import DatabaseManager

    db_manager = DatabaseManager()
    database_name = config.db_name

    if not await db_manager.database_exists(database_name):
        print(f"Database {database_name} does not exist. Creating...")
        await db_manager.create_database(database_name)
        print(f"Database {database_name} created successfully.")

    await db_manager.create_tables()

if __name__ == "__main__":
    import asyncio

    asyncio.run(main())



# metadata.py
""" MetaData model """
from datetime import datetime
from sqlalchemy import Column, String, DateTime, ForeignKey
from sqlalchemy.orm import relationship
import os
import sys
from ..database import Base


class MetaDatas(Base):
    """ MetaData model"""
    __tablename__ = "metadatas"

    id = Column(String, primary_key=True)
    user_id = Column(String, ForeignKey("users.id"), index=True)
    version = Column(String, nullable=False)
    contract_metadata = Column(String, nullable=False)
    memory_id = Column(String, ForeignKey("memories.id"), index=True)
    created_at = Column(DateTime, default=datetime.utcnow)
    updated_at = Column(DateTime, onupdate=datetime.utcnow)

    user = relationship("User", back_populates="metadatas")
    memory = relationship("MemoryModel", back_populates="metadatas")

    def __repr__(self):
        return f"<MetaData(id={self.id}, version={self.version}, field={self.field}, memory_id={self.memory_id}, created_at={self.created_at}, updated_at={self.updated_at})>"



import asyncio
# import logging
from typing import List
from qdrant_client import models
import instructor
from openai import OpenAI
from unstructured.cleaners.core import clean
from unstructured.partition.pdf import partition_pdf
from cognee.infrastructure.databases.vector.qdrant.QDrantAdapter import CollectionConfig
from cognee.infrastructure.llm.get_llm_client import get_llm_client
from cognee.modules.cognify.graph.add_classification_nodes import add_classification_nodes
from cognee.modules.cognify.graph.add_node_connections import add_node_connection, graph_ready_output, \
    connect_nodes_in_graph, extract_node_descriptions
from cognee.modules.cognify.graph.add_propositions import append_to_graph
from cognee.modules.cognify.llm.add_node_connection_embeddings import process_items
from cognee.modules.cognify.vector.batch_search import adapted_qdrant_batch_search
from cognee.modules.cognify.vector.add_propositions import add_propositions

from cognee.config import Config
from cognee.modules.cognify.llm.classify_content import classify_into_categories
from cognee.modules.cognify.llm.content_to_cog_layers import content_to_cog_layers
from cognee.modules.cognify.llm.generate_graph import generate_graph
from cognee.shared.data_models import DefaultContentPrediction,  KnowledgeGraph, DefaultCognitiveLayer
from cognee.infrastructure.databases.graph.get_graph_client import get_graph_client
from cognee.shared.data_models import GraphDBType
from cognee.infrastructure.databases.vector.get_vector_database import get_vector_database
from cognee.infrastructure.databases.relational import DuckDBAdapter
from cognee.modules.cognify.graph.add_document_node import add_document_node
from cognee.modules.cognify.graph.initialize_graph import initialize_graph

config = Config()
config.load()

aclient = instructor.patch(OpenAI())

USER_ID = "default_user"

async def cognify(dataset_name: str = "root"):
    """This function is responsible for the cognitive processing of the content."""

    db = DuckDBAdapter()
    files_metadata = db.get_files_metadata(dataset_name)

    awaitables = []

    await initialize_graph(USER_ID)

    for file_metadata in files_metadata:
        with open(file_metadata["file_path"], "rb") as file:
            elements = partition_pdf(file = file, strategy = "fast")
            text = "\n".join(map(lambda element: clean(element.text), elements))

            awaitables.append(process_text(text, file_metadata))

    graphs = await asyncio.gather(*awaitables)

    return graphs[0]

async def process_text(input_text: str, file_metadata: dict):
    print(f"Processing document ({file_metadata['id']})")
  
    classified_categories = []

    try:
        # Classify the content into categories
        classified_categories = await classify_into_categories(
            input_text,
            "classify_content.txt",
            DefaultContentPrediction
        )
        file_metadata["categories"] = list(map(lambda category: category["layer_name"], classified_categories))
    except Exception as e:
        print(e)
        raise e

    await add_document_node(f"DefaultGraphModel:{USER_ID}", file_metadata)
    print(f"Document ({file_metadata['id']}) categorized: {file_metadata['categories']}")

    cognitive_layers = await content_to_cog_layers(
        "generate_cog_layers.txt",
        classified_categories[0],
        response_model = DefaultCognitiveLayer
    )

    cognitive_layers = [layer_subgroup.name for layer_subgroup in cognitive_layers.cognitive_layers]

    async def generate_graph_per_layer(text_input: str, layers: List[str], response_model: KnowledgeGraph = KnowledgeGraph):
        generate_graphs_awaitables = [generate_graph(text_input, "generate_graph_prompt.txt", {"layer": layer}, response_model) for layer in
                layers]

        return await asyncio.gather(*generate_graphs_awaitables)

    # Run the async function for each set of cognitive layers
    layer_graphs = await generate_graph_per_layer(input_text, cognitive_layers)
    # print(layer_graphs)

    print(f"Document ({file_metadata['id']}) layer graphs created")

    # G = await create_semantic_graph(graph_model_instance)

    await add_classification_nodes(f"DOCUMENT:{file_metadata['id']}", classified_categories[0])

    unique_layer_uuids = await append_to_graph(layer_graphs, classified_categories[0])

    print(f"Document ({file_metadata['id']}) layers connected")

    graph_client = get_graph_client(GraphDBType.NETWORKX)

    await graph_client.load_graph_from_file()

    graph = graph_client.graph

    # # Extract the node descriptions
    node_descriptions = await extract_node_descriptions(graph.nodes(data = True))
    # print(node_descriptions)

    unique_layer_uuids = set(node["layer_decomposition_uuid"] for node in node_descriptions)

    collection_config = CollectionConfig(
        vector_config = {
            "content": models.VectorParams(
                distance = models.Distance.COSINE,
                size = 3072
            )
        },
    )

    try:
        for layer in unique_layer_uuids:
            db = get_vector_database()
            await db.create_collection(layer, collection_config)
    except Exception as e:
        print(e)

    await add_propositions(node_descriptions)

    grouped_data = await add_node_connection(node_descriptions)
    # print("we are here, grouped_data", grouped_data)

    llm_client = get_llm_client()

    relationship_dict = await process_items(grouped_data, unique_layer_uuids, llm_client)
    # print("we are here", relationship_dict[0])

    results = await adapted_qdrant_batch_search(relationship_dict, db)
    # print(results)

    relationship_d = graph_ready_output(results)
    # print(relationship_d)

    connect_nodes_in_graph(graph, relationship_d)

    print(f"Document ({file_metadata['id']}) processed")

    return graph



if __name__ == "__main__":
    asyncio.run(cognify("""In the nicest possible way, Britons have always been a bit silly about animals. “Keeping pets, for the English, is not so much a leisure activity as it is an entire way of life,” wrote the anthropologist Kate Fox in Watching the English, nearly 20 years ago. Our dogs, in particular, have been an acceptable outlet for emotions and impulses we otherwise keep strictly controlled – our latent desire to be demonstratively affectionate, to be silly and chat to strangers. If this seems like an exaggeration, consider the different reactions you’d get if you struck up a conversation with someone in a park with a dog, versus someone on the train.
Indeed, British society has been set up to accommodate these four-legged ambassadors. In the UK – unlike Australia, say, or New Zealand – dogs are not just permitted on public transport but often openly encouraged. Many pubs and shops display waggish signs, reading, “Dogs welcome, people tolerated”, and have treat jars on their counters. The other day, as I was waiting outside a cafe with a friend’s dog, the barista urged me to bring her inside.
For years, Britons’ non-partisan passion for animals has been consistent amid dwindling common ground. But lately, rather than bringing out the best in us, our relationship with dogs is increasingly revealing us at our worst – and our supposed “best friends” are paying the price.
As with so many latent traits in the national psyche, it all came unleashed with the pandemic, when many people thought they might as well make the most of all that time at home and in local parks with a dog. Between 2019 and 2022, the number of pet dogs in the UK rose from about nine million to 13 million. But there’s long been a seasonal surge around this time of year, substantial enough for the Dogs Trust charity to coin its famous slogan back in 1978: “A dog is for life, not just for Christmas.”
"""))


from typing import BinaryIO
from cognee.infrastructure.files import get_file_metadata, FileMetadata
from .IngestionData import IngestionData

def create_binary_data(data: BinaryIO):
    return BinaryData(data)

class BinaryData(IngestionData):
    data: BinaryIO = None
    metadata: FileMetadata = None

    def __init__(self, data: BinaryIO):
        self.data = data

    def get_identifier(self):
        self.ensure_metadata()

        return self.metadata["mime_type"] + "_" + "|".join(self.metadata["keywords"])

    def get_extension(self):
        self.ensure_metadata()

        return self.metadata["extension"]

    def ensure_metadata(self):
        if self.metadata is None:
            self.metadata = get_file_metadata(self.data)

    def get_data(self):
        return self.data



from typing import List
from abc import abstractmethod
from typing import Protocol

class GraphDBInterface(Protocol):

    """ Save and Load Graphs """

    @abstractmethod
    async def graph(self):
        raise NotImplementedError

    @abstractmethod
    async def save_graph_to_file(
        self,
        file_path: str = None
    ): raise NotImplementedError

    @abstractmethod
    async def load_graph_from_file(
        self,
        file_path: str = None
    ): raise NotImplementedError

    @abstractmethod
    async def delete_graph_from_file(
        self,
        path: str = None
    ): raise NotImplementedError

    """ CRUD operations on graph nodes """

    @abstractmethod

    async def add_node(
        self,
        id: str,
        **kwargs
    ): raise NotImplementedError

    @abstractmethod
    async def delete_node(
        self,
        id: str
    ): raise NotImplementedError


    """ CRUD operations on graph edges """


    @abstractmethod
    async def add_edge(
        self,
        from_node: str,
        to_node: str,
        **kwargs
    ): raise NotImplementedError


    # @abstractmethod
    # async def create_vector_index(
    #     self,
    #     collection_name: str,
    #     vector_index_config: object
    # ): raise NotImplementedError

    # @abstractmethod
    # async def create_data_index(
    #     self,
    #     collection_name: str,
    #     vector_index_config: object
    # ): raise NotImplementedError

    # """ Data points """
    # @abstractmethod
    # async def create_data_points(
    #     self,
    #     collection_name: str,
    #     data_points: List[any]
    # ): raise NotImplementedError

    # @abstractmethod
    # async def get_data_point(
    #     self,
    #     collection_name: str,
    #     data_point_id: str
    # ): raise NotImplementedError

    # @abstractmethod
    # async def update_data_point(
    #     self,
    #     collection_name: str,
    #     data_point_id: str,
    #     payload: object
    # ): raise NotImplementedError

    # @abstractmethod
    # async def delete_data_point(
    #     self,
    #     collection_name: str,
    #     data_point_id: str
    # ): raise NotImplementedError



""" This module is responsible for creating a semantic graph """
from typing import  Optional, Any
from pydantic import BaseModel
from cognee.infrastructure.databases.graph.get_graph_client import get_graph_client
from cognee.shared.data_models import GraphDBType


async def generate_node_id(instance: BaseModel) -> str:
    for field in ["id", "doc_id", "location_id", "type_id"]:
        if hasattr(instance, field):
            return f"{instance.__class__.__name__}:{getattr(instance, field)}"
    return f"{instance.__class__.__name__}:default"

async def add_node_and_edge(client, parent_id: Optional[str], node_id: str, node_data: dict, relationship_data: dict):
    await client.add_node(node_id, **node_data)  # Add the current node with its data
    if parent_id:
        # Add an edge between the parent node and the current node with the correct relationship data
        await client.add_edge(parent_id, node_id, **relationship_data)


async def process_attribute(graph_client, parent_id: Optional[str], attribute: str, value: Any):
    if isinstance(value, BaseModel):
        node_id = await generate_node_id(value)

        node_data = value.dict(exclude={"default_relationship"})

        # Use the specified default relationship for the edge between the parent node and the current node
        relationship_data = value.default_relationship.dict() if hasattr(value, "default_relationship") else {}

        await add_node_and_edge(graph_client, parent_id, node_id, node_data, relationship_data)

        # Recursively process nested attributes to ensure all nodes and relationships are added to the graph
        for sub_attr, sub_val in value.__dict__.items():  # Access attributes and their values directly
            await process_attribute(graph_client, node_id, sub_attr, sub_val)

    elif isinstance(value, list) and all(isinstance(item, BaseModel) for item in value):
        # For lists of BaseModel instances, process each item in the list
        for item in value:
            await process_attribute(graph_client, parent_id, attribute, item)

async def create_dynamic(graph_model) :
    root_id = await generate_node_id(graph_model)

    node_data = graph_model.dict(exclude = {"default_relationship", "id"})

    graph_client = get_graph_client(GraphDBType.NETWORKX)

    await graph_client.add_node(root_id, **node_data)

    for attribute_name, attribute_value in graph_model:
        await process_attribute(graph_client, root_id, attribute_name, attribute_value)

    return graph_client


async def create_semantic_graph(graph_model_instance):
    # Dynamic graph creation based on the provided graph model instance
    graph = await create_dynamic(graph_model_instance)

    return graph



# if __name__ == "__main__":
#     import asyncio

#     # Assuming all necessary imports and GraphDBType, get_graph_client, Document, DocumentType, etc. are defined

#     # Initialize the graph client
#     graph_client = get_graph_client(GraphDBType.NETWORKX)

#     # Define a GraphModel instance with example data
#     graph_model_instance = DefaultGraphModel(
#         id="user123",
#         documents=[
#             Document(
#                 doc_id="doc1",
#                 title="Document 1",
#                 summary="Summary of Document 1",
#                 content_id="content_id_for_doc1",
#                 doc_type=DocumentType(type_id="PDF", description="Portable Document Format"),
#                 categories=[
#                     Category(category_id="finance", name="Finance", default_relationship=Relationship(type="belongs_to")),
#                     Category(category_id="tech", name="Technology", default_relationship=Relationship(type="belongs_to"))
#                 ],
#                 default_relationship=Relationship(type="has_document")
#             ),
#             Document(
#                 doc_id="doc2",
#                 title="Document 2",
#                 summary="Summary of Document 2",
#                 content_id="content_id_for_doc2",
#                 doc_type=DocumentType(type_id="TXT", description="Text File"),
#                 categories=[
#                     Category(category_id="health", name="Health", default_relationship=Relationship(type="belongs_to")),
#                     Category(category_id="wellness", name="Wellness", default_relationship=Relationship(type="belongs_to"))
#                 ],
#                 default_relationship=Relationship(type="has_document")
#             )
#         ],
#         user_properties=UserProperties(
#             custom_properties={"age": "30"},
#             location=UserLocation(location_id="ny", description="New York", default_relationship=Relationship(type="located_in"))
#         ),
#         default_fields={
#             "created_at": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
#             "updated_at": datetime.now().strftime("%Y-%m-%d %H:%M:%S")
#         }
#     )

#     # Run the graph creation asynchronously
#     G = asyncio.run(create_semantic_graph(graph_model_instance, graph_client))

#     # Optionally, here you can add more nodes, edges, or perform other operations on the graph G

# async def create_semantic_graph(
# ):
#     graph_type = GraphDBType.NETWORKX
#
#     # Call the get_graph_client function with the selected graph type
#     graph_client = get_graph_client(graph_type)
#
#     print(graph_client)
#
#     await graph_client.load_graph_from_file()
#     #
#     #
#     #
#     # b = await graph_client.add_node("23ds",     {
#     #     "username": "exampleUser",
#     #     "email": "user@example.com"
#     # })
#     #
#     # await graph_client.save_graph_to_file(b)
#     graph_model_instance = DefaultGraphModel(
#         id="user123",
#         documents=[
#             Document(
#                 doc_id="doc1",
#                 title="Document 1",
#                 summary="Summary of Document 1",
#                 content_id="content_id_for_doc1",  # Assuming external content storage ID
#                 doc_type=DocumentType(type_id="PDF", description="Portable Document Format"),
#                 categories=[
#                     Category(category_id="finance", name="Finance",
#                              default_relationship=Relationship(type="belongs_to")),
#                     Category(category_id="tech", name="Technology",
#                              default_relationship=Relationship(type="belongs_to"))
#                 ],
#                 default_relationship=Relationship(type='has_document')
#             ),
#             Document(
#                 doc_id="doc2",
#                 title="Document 2",
#                 summary="Summary of Document 2",
#                 content_id="content_id_for_doc2",
#                 doc_type=DocumentType(type_id="TXT", description="Text File"),
#                 categories=[
#                     Category(category_id="health", name="Health", default_relationship=Relationship(type="belongs_to")),
#                     Category(category_id="wellness", name="Wellness",
#                              default_relationship=Relationship(type="belongs_to"))
#                 ],
#                 default_relationship=Relationship(type='has_document')
#             )
#         ],
#         user_properties=UserProperties(
#             custom_properties={"age": "30"},
#             location=UserLocation(location_id="ny", description="New York",
#                                   default_relationship=Relationship(type='located_in'))
#         ),
#         default_fields={
#             'created_at': datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
#             'updated_at': datetime.now().strftime("%Y-%m-%d %H:%M:%S")
#         }
#     )
#
#     G = await create_dynamic(graph_model_instance, graph_client)
#
#     # print("Nodes and their data:")
#     # for node, data in G.graph.nodes(data=True):
#     #     print(node, data)
#     #
#     # # Print edges with their data
#     # print("\nEdges and their data:")
#     # for source, target, data in G.graph.edges(data=True):
#     #     print(f"{source} -> {target} {data}")
#     # print(G)
#
#
#
#
#
#
#
#
#
#
#     # return await graph_client.create( user_id = user_id, custom_user_properties=custom_user_properties, required_layers=required_layers, default_fields=default_fields, existing_graph=existing_graph)
#
#
# if __name__ == "__main__":
#     import asyncio
#
#     user_id = 'user123'
#     custom_user_properties = {
#         'username': 'exampleUser',
#         'email': 'user@example.com'
#     }
#     asyncio.run(create_semantic_graph())


""" This module contains the code to classify content into categories using the LLM API. """
from typing import Type, List
from pydantic import BaseModel
from cognee.infrastructure.llm.get_llm_client import get_llm_client
from cognee.utils import read_query_prompt

async def classify_into_categories(text_input: str, system_prompt_path: str, response_model: Type[BaseModel]):
    llm_client = get_llm_client()

    system_prompt = await read_query_prompt(system_prompt_path)

    llm_output = await llm_client.acreate_structured_output(text_input, system_prompt, response_model)

    return extract_categories(llm_output.dict())

def extract_categories(llm_output) -> List[dict]:
    # Extract the first subclass from the list (assuming there could be more)
    layer_enum = llm_output["label"]["subclass"][0]

    # The data type is derived from "type" and converted to lowercase
    data_type = llm_output["label"]["type"].lower()

    # The layer name is the name of the Enum member (e.g., "NEWS_STORIES")
    # layer_name = layer_enum.name.replace("_", " ").title()

    # The layer name is the value of the Enum member (e.g., "News stories and blog posts")
    layer_name = layer_enum.value

    return [{
        "data_type": data_type,
        "layer_name": layer_name  # llm layer classification
    }]

# if __name__ == "__main__":
#     import asyncio
#     asyncio.run(classify_into_categories("""Russia summons US ambassador in Moscow and says it will expel diplomats who meddle in its internal affairs
# The Russian foreign ministry said on Thursday it had summoned the US ambassador in Moscow and warned her against “attempts to interfere in the internal affairs of the Russian Federation”, reports Reuters.

# Ahead of a March presidential election, it said in a statement that such behaviour would be “firmly and resolutely suppressed, up to and including the expulsion as ‘persona non grata’ of US embassy staff involved in such actions”.""", "classify_content.txt", ContentPrediction))




""" This module contains the classifiers for the documents. """

import json
import logging

from langchain.prompts import ChatPromptTemplate
from langchain.document_loaders import TextLoader
from langchain.document_loaders import DirectoryLoader
from langchain.chains import create_extraction_chain
from langchain.chat_models import ChatOpenAI

from ..config import Config

config = Config()
config.load()
OPENAI_API_KEY = config.openai_key

async def classify_documents(query: str, document_id: str, content: str):
    """Classify the documents based on the query and content."""
    document_context = content
    logging.info("This is the document context %s", document_context)

    llm = ChatOpenAI(temperature=0, model=config.model)
    prompt_classify = ChatPromptTemplate.from_template(
        """You are a summarizer and classifier.
         Determine what book this is and where does it belong in the output :
          {query}, Id: {d_id} Document context is: {context}"""
    )
    json_structure = [
        {
            "name": "summarizer",
            "description": "Summarization and classification",
            "parameters": {
                "type": "object",
                "properties": {
                    "DocumentCategory": {
                        "type": "string",
                        "description": "The classification of documents "
                                       "in groups such as legal, medical, etc.",
                    },
                    "Title": {
                        "type": "string",
                        "description": "The title of the document",
                    },
                    "Summary": {
                        "type": "string",
                        "description": "The summary of the document",
                    },
                    "d_id": {"type": "string", "description": "The id of the document"},
                },
                "required": ["DocumentCategory", "Title", "Summary", "d_id"],
            },
        }
    ]
    chain_filter = prompt_classify | llm.bind(
        function_call={"name": "summarizer"}, functions=json_structure
    )
    classifier_output = await chain_filter.ainvoke(
        {"query": query, "d_id": document_id, "context": str(document_context)}
    )

    arguments_str = classifier_output.additional_kwargs["function_call"]["arguments"]
    logging.info("This is the arguments string %s", arguments_str)
    arguments_dict = json.loads(arguments_str)
    return arguments_dict



""" This module contains the search function that is used to search for nodes in the graph."""
import asyncio
from enum import Enum, auto
from typing import Dict, Any, Callable, List

from cognee.modules.search.graph.search_adjacent import search_adjacent
from cognee.modules.search.vector.search_similarity import search_similarity
from cognee.modules.search.graph.search_categories import search_categories
from cognee.modules.search.graph.search_neighbour import search_neighbour


class SearchType(Enum):
    ADJACENT = auto()
    SIMILARITY = auto()
    CATEGORIES = auto()
    NEIGHBOR = auto()


async def search(graph, query_params: Dict[SearchType, Dict[str, Any]]) -> List:
    search_functions: Dict[SearchType, Callable] = {
        SearchType.ADJACENT: search_adjacent,
        SearchType.SIMILARITY: search_similarity,
        SearchType.CATEGORIES: search_categories,
        SearchType.NEIGHBOR: search_neighbour,
    }

    results = []

    # Create a list to hold all the coroutine objects
    search_tasks = []

    for search_type, params in query_params.items():
        search_func = search_functions.get(search_type)
        if search_func:
            # Schedule the coroutine for execution and store the task
            full_params = {**params, 'graph': graph}
            task = search_func(**full_params)
            search_tasks.append(task)

    # Use asyncio.gather to run all scheduled tasks concurrently
    search_results = await asyncio.gather(*search_tasks)

    # Update the results set with the results from all tasks
    for search_result in search_results:
        results.append(search_result)

    return results

# if __name__ == "__main__":
#     import asyncio

#     query_params = {
#         SearchType.SIMILARITY: {'query': 'your search query here'}
#     }
#     async def main():
#         graph_client = get_graph_client(GraphDBType.NETWORKX)

#         await graph_client.load_graph_from_file()
#         graph = graph_client.graph
#         results = await search(graph, query_params)
#         print(results)

#     asyncio.run(main())



from .job import Job


class Response:
    def __init__(
        self,
        error=None,
        message=None,
        successful_uploads=None,
        failed_uploads=None,
        empty_files_count=None,
        duplicate_files_count=None,
        job_id=None,
        jobs=None,
        job_status=None,
        status_code=None,
    ):
        self.error = error
        self.message = message
        self.successful_uploads = successful_uploads
        self.failed_uploads = failed_uploads
        self.empty_files_count = empty_files_count
        self.duplicate_files_count = duplicate_files_count
        self.job_id = job_id
        self.jobs = jobs
        self.job_status = job_status
        self.status_code = status_code

    @classmethod
    def from_json(cls, json_dict, status_code):
        successful_uploads = cls._convert_successful_uploads_to_jobs(
            json_dict.get("successful_uploads", None)
        )
        jobs = cls._convert_to_jobs(json_dict.get("Jobs", None))

        return cls(
            error=json_dict.get("error"),
            message=json_dict.get("message"),
            successful_uploads=successful_uploads,
            failed_uploads=json_dict.get("failed_uploads"),
            empty_files_count=json_dict.get("empty_files_count"),
            duplicate_files_count=json_dict.get("duplicate_files_count"),
            job_id=json_dict.get("JobID"),
            jobs=jobs,
            job_status=json_dict.get("JobStatus"),
            status_code=status_code,
        )

    @classmethod
    def _convert_successful_uploads_to_jobs(cls, successful_uploads):
        if not successful_uploads:
            return None
        return [
            Job(filename=key, job_id=val) for key, val in successful_uploads.items()
        ]

    @classmethod
    def _convert_to_jobs(cls, jobs):
        if not jobs:
            return None
        return [Job(job_id=job["JobID"], job_status=job["JobStatus"]) for job in jobs]

    def __str__(self):
        attributes = []
        if self.error is not None:
            attributes.append(f"error: {self.error}")
        if self.message is not None:
            attributes.append(f"message: {self.message}")
        if self.successful_uploads is not None:
            attributes.append(f"successful_uploads: {str(self.successful_uploads)}")
        if self.failed_uploads is not None:
            attributes.append(f"failed_uploads: {self.failed_uploads}")
        if self.empty_files_count is not None:
            attributes.append(f"empty_files_count: {self.empty_files_count}")
        if self.duplicate_files_count is not None:
            attributes.append(f"duplicate_files_count: {self.duplicate_files_count}")
        if self.job_id is not None:
            attributes.append(f"job_id: {self.job_id}")
        if self.jobs is not None:
            attributes.append(f"jobs: {str(self.jobs)}")
        if self.job_status is not None:
            attributes.append(f"job_status: {self.job_status}")
        if self.status_code is not None:
            attributes.append(f"status_code: {self.status_code}")

        return "Response(" + ", ".join(attributes) + ")"



import duckdb
from cognee.root_dir import get_absolute_path

class DuckDBAdapter():
    def __init__(self):
        db_path = get_absolute_path("./data/cognee")
        db_location = db_path + "/cognee.duckdb"

        self.db_client = duckdb.connect(db_location)

    def get_datasets(self):
        tables = self.db_client.sql("SELECT DISTINCT schema_name FROM duckdb_tables();").to_df().to_dict("list")

        return list(
            filter(
                lambda table_name: table_name.endswith('staging') is False,
                tables["schema_name"]
            )
        )

    def get_files_metadata(self, dataset_name: str):
        return self.db_client.sql(f"SELECT * FROM {dataset_name}.file_metadata;").to_df().to_dict("records")



from .IngestionData import IngestionData

def create_text_data(data: str):
    return TextData(data)

class TextData(IngestionData):
    data: str = None

    def __init__(self, data: str):
        self.data = data

    def get_data(self):
        return self.data

    def get_chunks(self):
        pass



import asyncio

async def process_items(grouped_data, unique_layer_uuids, llm_client):
    results_to_check = []  # This will hold results excluding self comparisons
    tasks = []  # List to hold all tasks
    task_to_info = {}  # Dictionary to map tasks to their corresponding group id and item info

    # Iterate through each group in grouped_data
    for group_id, items in grouped_data.items():
        # Filter unique_layer_uuids to exclude the current group_id
        target_uuids = [uuid for uuid in unique_layer_uuids if uuid != group_id]

        # Process each item in the group
        for item in items:
            # For each target UUID, create an async task for the item's embedding retrieval
            for target_id in target_uuids:
                task = asyncio.create_task(llm_client.async_get_embedding_with_backoff(item['description'], "text-embedding-3-large"))
                tasks.append(task)
                # Map the task to the target id, item's node_id, and description for later retrieval
                task_to_info[task] = (target_id, item['node_id'], group_id, item['description'])

    # Await all tasks to complete and gather results
    results = await asyncio.gather(*tasks)

    # Process the results, associating them with their target id, node id, and description
    for task, embedding in zip(tasks, results):
        target_id, node_id, group_id, description = task_to_info[task]
        results_to_check.append([target_id, embedding, description, node_id, group_id])

    return results_to_check



""" This module contains the MemoryModel class, which is a SQLAlchemy model for the memory table in the relational database. """
from datetime import datetime
from sqlalchemy import Column, String, DateTime, ForeignKey, UUID
from sqlalchemy.orm import relationship
from ..database import Base


class MemoryModel(Base):
    """ Memory model"""
    __tablename__ = "memories"

    id = Column(UUID, primary_key=True)
    user_id = Column(String, ForeignKey("users.id"), index=True)
    operation_id = Column(String, ForeignKey("operations.id"), index=True)
    memory_name = Column(String, nullable=True)
    memory_category = Column(String, nullable=True)
    created_at = Column(DateTime, default=datetime.utcnow)
    updated_at = Column(DateTime, onupdate=datetime.utcnow)
    methods_list = Column(String, nullable=True)
    attributes_list = Column(String, nullable=True)

    user = relationship("User", back_populates="memories")
    operation = relationship("Operation", back_populates="memories")
    metadatas = relationship(
        "MetaDatas", back_populates="memory", cascade="all, delete-orphan"
    )

    def __repr__(self):
        return f"<Memory(id={self.id}, user_id={self.user_id}, created_at={self.created_at}, updated_at={self.updated_at})>"



"""This module provides functionalities for creating and managing databases."""

import os
from contextlib import asynccontextmanager
from sqlalchemy import text
from sqlalchemy.ext.asyncio import create_async_engine
from cognee.config import Config
from cognee.database.relationaldb.database import Base, get_sqlalchemy_database_url

globalConfig = Config()

class DatabaseManager:
    """Manages database creation, deletion, and table initialization."""
    def __init__(self):
        """Initialize the Database Url with a given configuration."""
        self.engine = create_async_engine(get_sqlalchemy_database_url(globalConfig.db_type), echo = True)
        self.db_type = globalConfig.db_type

    @asynccontextmanager
    async def get_connection(self):
        """Initialize the DatabaseManager with a given configuration."""
        if self.db_type in ["sqlite", "duckdb"]:
            # For SQLite and DuckDB, the engine itself manages connections
            yield self.engine
        else:
            async with self.engine.connect() as connection:
                yield connection

    async def database_exists(self, db_name):
        """Check if a database exists."""
        if self.db_type in ["sqlite", "duckdb"]:
            # For SQLite and DuckDB, check if the database file exists
            return os.path.exists(db_name)
        else:
            query = text(f"SELECT 1 FROM pg_database WHERE datname='{db_name}'")
            async with self.get_connection() as connection:
                result = await connection.execute(query)
                return  result.fetchone() is not None

    async def create_database(self, db_name):
        """Create a new database."""
        if self.db_type not in ["sqlite", "duckdb"]:
            # For databases like PostgreSQL, create the database explicitly
            async with self.get_connection() as connection:
                await connection.execute(text(f"CREATE DATABASE {db_name}"))

    async def drop_database(self, db_name):
        """Drop an existing database."""
        if self.db_type in ["sqlite", "duckdb"]:
            # For SQLite and DuckDB, simply remove the database file
            os.remove(db_name)
        else:
            async with self.get_connection() as connection:
                await connection.execute(text(f"DROP DATABASE IF EXISTS {db_name}"))

    async def create_tables(self):
        """Create tables based on the SQLAlchemy Base metadata."""
        try:
            async with self.engine.begin() as conn:
                await conn.run_sync(Base.metadata.create_all)
        except Exception as e:
            print(e)
            raise e



from .embeddings_type import EmbeddingsType
from ..chunkers.chunk_strategy import ChunkStrategy


class Embeddings:
    def __init__(
        self,
        embeddings_type: EmbeddingsType = EmbeddingsType.OPEN_AI,
        chunk_size: int = 256,
        chunk_overlap: int = 128,
        chunk_strategy: ChunkStrategy = ChunkStrategy.EXACT,
        docker_image: str = None,
        hugging_face_model_name: str = None,
    ):
        self.embeddings_type = embeddings_type
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
        self.chunk_strategy = chunk_strategy
        self.docker_image = docker_image
        self.hugging_face_model_name = hugging_face_model_name

    def serialize(self):
        data = {
            "embeddings_type": self.embeddings_type.name
            if self.embeddings_type
            else None,
            "chunk_size": self.chunk_size,
            "chunk_overlap": self.chunk_overlap,
            "chunk_strategy": self.chunk_strategy.name if self.chunk_strategy else None,
            "docker_image": self.docker_image,
            "hugging_face_model_name": self.hugging_face_model_name,
        }

        return {k: v for k, v in data.items() if v is not None}



from typing import List, Optional, Dict
from pydantic import BaseModel, Field
from qdrant_client import AsyncQdrantClient, models
from ..vector_db_interface import VectorDBInterface

class CollectionConfig(BaseModel, extra = "forbid"):
    vector_config: Dict[str, models.VectorParams] = Field(..., description="Vectors configuration" )
    hnsw_config: Optional[models.HnswConfig] = Field(default = None, description="HNSW vector index configuration")
    optimizers_config: Optional[models.OptimizersConfig] = Field(default = None, description="Optimizers configuration")
    quantization_config: Optional[models.QuantizationConfig] = Field(default = None, description="Quantization configuration")

class QDrantAdapter(VectorDBInterface):
    qdrant_url: str = None
    qdrant_path: str = None
    qdrant_api_key: str = None
  
    def __init__(self, qdrant_path, qdrant_url, qdrant_api_key):
        if qdrant_path is not None:
            self.qdrant_path = qdrant_path
        else:
            self.qdrant_url = qdrant_url

        self.qdrant_api_key = qdrant_api_key

    def get_qdrant_client(self) -> AsyncQdrantClient:
        if self.qdrant_path is not None:
            return AsyncQdrantClient(
                path = self.qdrant_path,
            )
        elif self.qdrant_url is not None:
            return AsyncQdrantClient(
                url = self.qdrant_url,
                api_key = self.qdrant_api_key,
            )

        return AsyncQdrantClient(
            location = ":memory:"
        )

    async def create_collection(
      self,
      collection_name: str,
      collection_config: CollectionConfig,
    ):
        client = self.get_qdrant_client()

        return await client.create_collection(
            collection_name = collection_name,
            vectors_config = collection_config.vector_config,
            hnsw_config = collection_config.hnsw_config,
            optimizers_config = collection_config.optimizers_config,
            quantization_config = collection_config.quantization_config
        )

    async def create_data_points(self, collection_name: str, data_points):
        client = self.get_qdrant_client()

        return await client.upload_points(
            collection_name = collection_name,
            points = data_points
        )

    async def search(self, collection_name: str, query_vector: List[float], limit: int, with_vector: bool = False):
        client = self.get_qdrant_client()

        return await client.search(
            collection_name = collection_name,
            query_vector = (
            "content", query_vector),
            limit = limit,
            with_vectors = with_vector
        )


    async def batch_search(self, collection_name: str, embeddings: List[List[float]],
                                  with_vectors: List[bool] = None):
        """
        Perform batch search in a Qdrant collection with dynamic search requests.

        Args:
        - collection_name (str): Name of the collection to search in.
        - embeddings (List[List[float]]): List of embeddings to search for.
        - limits (List[int]): List of result limits for each search request.
        - with_vectors (List[bool], optional): List indicating whether to return vectors for each search request.
            Defaults to None, in which case vectors are not returned.

        Returns:
        - results: The search results from Qdrant.
        """

        client = self.get_qdrant_client()

        # Default with_vectors to False for each request if not provided
        if with_vectors is None:
            with_vectors = [False] * len(embeddings)

        # Ensure with_vectors list matches the length of embeddings and limits
        if len(with_vectors) != len(embeddings):
            raise ValueError("The length of with_vectors must match the length of embeddings and limits")

        # Generate dynamic search requests based on the provided embeddings
        requests = [
            models.SearchRequest(vector=models.NamedVector(
                name="content",
                vector=embedding,
            ),
                # vector= embedding,
                limit=3,
                with_vector=False
            ) for embedding in [embeddings]
        ]

        # Perform batch search with the dynamically generated requests
        results = await client.search_batch(
            collection_name=collection_name,
            requests=requests
        )

        return results
